@@ -8,6 +8,9 @@ class VectorImage:
     def __init__(self, source):
         self.source = source
 
+    def __str__(self):
+        return self.source
+    
     def display(self):
         process = Popen("rsvg-convert -x 4 -y 4 | viu -", shell=True, stdin=PIPE, stdout=stdout)
         process.communicate(input=str.encode(self.source))
