@@ -1,51 +1,52 @@
 import json
-from pyDigitalWaveTools.vcd.parser import VcdParser
+# from pyDigitalWaveTools.vcd.parser import VcdParser
+from vcd.reader import *
 
-from .exceptions import SoottyInternalError
+from .exceptions import *
 from .limits import LimitExpression
 
 
 class Wire:
 
-    def __init__(self, name, width=1, data=[]):
+    def __init__(self, name, width=1, data=None):
         self.name = name
         self.width = width
-        self.data = data
+        self.data = list() if data is None else data
     
-    @staticmethod
-    def _to_int(data):
-        if data[0] == 'b':
-            try:
-                return int(data[1:], 2)
-            except ValueError:
-                return None
-        else:
-            try:
-                return int(data)
-            except ValueError:
-                return None
+    # @staticmethod
+    # def _to_int(data):
+    #     if data[0] == 'b':
+    #         try:
+    #             return int(data[1:], 2)
+    #         except ValueError:
+    #             return None
+    #     else:
+    #         try:
+    #             return int(data)
+    #         except ValueError:
+    #             return None
 
-    @staticmethod
-    def from_vcd(vcd_data, base_name=""):
-        wiredata = []
-        source_data = sorted(vcd_data["data"])
-        source_dict = dict(source_data)
-        time = 0
-        end = source_data[-1][0] if len(source_data) else 0
-        while time <= end:
-            if time in source_dict:
-                wiredata.append(Wire._to_int(source_dict[time]))
-            elif time > 0:
-                wiredata.append(wiredata[-1])
-            else:
-                wiredata.append(None)
-            time += 1
+    # @staticmethod
+    # def from_vcd(vcd_data, base_name=""):
+    #     wiredata = []
+    #     source_data = sorted(vcd_data["data"])
+    #     source_dict = dict(source_data)
+    #     time = 0
+    #     end = source_data[-1][0] if len(source_data) else 0
+    #     while time <= end:
+    #         if time in source_dict:
+    #             wiredata.append(Wire._to_int(source_dict[time]))
+    #         elif time > 0:
+    #             wiredata.append(wiredata[-1])
+    #         else:
+    #             wiredata.append(None)
+    #         time += 1
         
-        return Wire(
-            name=base_name + ('.' if len(base_name) else "") + vcd_data["name"],
-            width=vcd_data["type"]["width"],
-            data=wiredata
-        )
+    #     return Wire(
+    #         name=base_name + ('.' if len(base_name) else "") + vcd_data["name"],
+    #         width=vcd_data["type"]["width"],
+    #         data=wiredata
+    #     )
 
     def __invert__(self):
         data = []
@@ -353,22 +354,22 @@ class WireGroup:
         for wire in wiregroup.wires:
             self.add_wire(wire)
 
-    @staticmethod
-    def from_vcd(vcd_data, base_name=None):
-        name = "" if base_name == None else (base_name + ('.' if len(base_name) else "") + vcd_data["name"])
-        wiregroup = WireGroup(
-            name=vcd_data["name"] if len(name) == 0 else name
-        )
-        for child in vcd_data["children"]:
-            if "data" in child:
-                wiregroup.add_wire(
-                    Wire.from_vcd(child, name)
-                )
-            else:
-                wiregroup.add_wires(
-                    WireGroup.from_vcd(child, name)
-                )
-        return wiregroup
+    # @staticmethod
+    # def from_vcd(vcd_data, base_name=None):
+    #     name = "" if base_name == None else (base_name + ('.' if len(base_name) else "") + vcd_data["name"])
+    #     wiregroup = WireGroup(
+    #         name=vcd_data["name"] if len(name) == 0 else name
+    #     )
+    #     for child in vcd_data["children"]:
+    #         if "data" in child:
+    #             wiregroup.add_wire(
+    #                 Wire.from_vcd(child, name)
+    #             )
+    #         else:
+    #             wiregroup.add_wires(
+    #                 WireGroup.from_vcd(child, name)
+    #             )
+    #     return wiregroup
 
 
 class WireTrace:
@@ -379,21 +380,145 @@ class WireTrace:
     def add_wiregroup(self, wiregroup):
         self.wiregroups.append(wiregroup)
 
-    @staticmethod
-    def read_vcd(filename):
-        with open(filename) as vcd_file:
-            vcd = VcdParser()
-            vcd.parse(vcd_file)
-            return vcd.scope.toJson()  # dump json here for debugging
+    # @staticmethod
+    # def read_vcd(filename):
+    #     with open(filename) as vcd_file:
+    #         vcd = VcdParser()
+    #         vcd.parse(vcd_file)
+    #         return vcd.scope.toJson()  # dump json here for debugging
         
+    # @staticmethod
+    # def from_vcd_file(filename):
+    #     vcd_data = WireTrace.read_vcd(filename)  # Read vcd data from file
+    #     wiretrace = WireTrace()
+    #     for child in vcd_data["children"]:
+    #         wiretrace.add_wiregroup(
+    #             WireGroup.from_vcd(child)
+    #         )
+    #     return wiretrace
+
     @staticmethod
-    def from_vcd_file(filename):
-        vcd_data = WireTrace.read_vcd(filename)  # Read vcd data from file
+    def from_vcd(filename):
+        """
+        Construct a WireTrace object from a parsed vcd file, using the pyvcd library.
+
+        Syntax of four-state VCD file (IEEE 1364-2005 §18.2.1):
+
+        value_change_dump_definitions ::= 
+            { declaration_command }{ simulation_command }
+        declaration_command ::= 
+            declaration_keyword 
+            [ command_text ] 
+            $end
+        simulation_command ::=
+            simulation_keyword { value_change } $end
+            | $comment [ comment_text ] $end
+            | simulation_time
+            | value_change
+        declaration_keyword ::=
+            $comment | $date | $enddefinitions | $scope | $timescale | $upscope
+            | $var | $version
+        simulation_keyword ::=
+            $dumpall | $dumpoff | $dumpon | $dumpvars
+        simulation_time ::=
+            # decimal_number
+        value_change ::=
+            scalar_value_change
+            | vector_value_change
+        scalar_value_change ::=
+            value identifier_code
+        value ::=
+            0 | 1 | x | X | z | Z
+        vector_value_change ::=
+            b binary_number identifier_code
+            | B binary_number identifier_code
+            | r real_number identifier_code
+            | R real_number identifier_code
+        identifier_code ::=
+            { ASCII character }
+        """
+
+        this = WireTrace()
+        this.metadata = dict()  # dictionary of vcd metadata
+        wires = dict()  # map from id_code to wire object
+
+        with open(filename, 'rb') as stream:
+            tokens = tokenize(stream)
+            for token in tokens:
+                if token.kind is TokenKind.COMMENT:
+                    this.metadata['comment'] = token.comment
+                elif token.kind is TokenKind.DATE:
+                    this.metadata['date'] = token.date
+                elif token.kind is TokenKind.ENDDEFINITIONS:
+                    break  # end of definitions
+                elif token.kind is TokenKind.SCOPE:
+                    pass  # implement scoping
+                elif token.kind is TokenKind.TIMESCALE:
+                    this.metadata['timescale'] = token.timescale
+                elif token.kind is TokenKind.UPSCOPE:
+                    pass  # implement scoping
+                elif token.kind is TokenKind.VAR:
+                    wires[token.var.id_code] = Wire(
+                        name=token.var.reference,
+                        width=token.var.size,
+                    )
+                elif token.kind is TokenKind.VERSION:
+                    this.metadata['version'] = token.version
+                else:
+                    raise SoottyError(f'Invalid vcd token when parsing: {token}')
+        
+            time = None
+            for token in tokens:
+                if token.kind is TokenKind.CHANGE_TIME:
+                    time = token.time_change
+                elif token.kind is TokenKind.CHANGE_SCALAR:
+                    value = token.scalar_change.value
+                    value = int(value) if value in ('0', '1') else value
+                    wires[token.scalar_change.id_code].data.append(value)  # TODO: fix to use time
+                elif token.kind is TokenKind.CHANGE_VECTOR:
+                    value = token.vector_change.value
+                    wires[token.vector_change.id_code].data.append(value)  # TODO: fix to use time
+                elif token.kind is TokenKind.CHANGE_REAL:
+                    print(token.real_change)
+                    raise SoottyInternalError(f'You forgot to implement token CHANGE_REAL.')
+                elif token.kind is TokenKind.CHANGE_STRING:
+                    print(token.string_change)
+                    raise SoottyInternalError(f'You forgot to implement token CHANGE_STRING.')
+                elif token.kind is TokenKind.DUMPALL:
+                    pass  # not sure what to do here
+                elif token.kind is TokenKind.DUMPOFF:
+                    pass  # not sure what to do here
+                elif token.kind is TokenKind.DUMPON:
+                    pass  # not sure what to do here
+                elif token.kind is TokenKind.DUMPVARS:
+                    pass  # not sure what to do here
+                elif token.kind is TokenKind.END:
+                    pass  # not sure what to do here
+                else:
+                    raise SoottyError(f'Invalid vcd token when parsing: {token}')
+
+            group = WireGroup("")
+            for wire in wires:
+                group.add_wire(wires[wire])
+            this.add_wiregroup(group)
+
+            return this
+
+    @staticmethod
+    def from_pyrtl(sim_trace):
+        """Parses a WireTrace object from a PyRTL SimulationTrace object.
+
+        :param SimulationTrace sim_trace: The object that stores the PyRTL tracer.
+        """
         wiretrace = WireTrace()
-        for child in vcd_data["children"]:
-            wiretrace.add_wiregroup(
-                WireGroup.from_vcd(child)
-            )
+        wiregroup = WireGroup('main')
+        for wirename in sim_trace.trace:
+            wiregroup.add_wire(Wire(
+                name = wirename,
+                width = sim_trace._wires[wirename].bitwidth,
+                data = sim_trace.trace[wirename]
+            ))
+        wiretrace.add_wiregroup(wiregroup)
         return wiretrace
 
     # Returns the time of the last entry in data
